@@ -1,5 +1,6 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- *  Copyright (C) 2002, 2003 Marco Pesenti Gritti
+ *  Copyright © 2002, 2003 Marco Pesenti Gritti
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -13,40 +14,43 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- *  $Id: ephy-history.c,v 1.48.10.2 2005/01/30 20:12:07 chpe Exp $
+ *  $Id: ephy-history.c 7981 2008-02-25 19:37:26Z chpe $
  */
 
-#ifdef HAVE_CONFIG_H
 #include "config.h"
-#endif
 
 #include "ephy-history.h"
+#include "ephy-marshal.h"
 #include "ephy-file-helpers.h"
 #include "ephy-debug.h"
 #include "ephy-node-db.h"
 #include "ephy-node-common.h"
 #include "eel-gconf-extensions.h"
 #include "ephy-prefs.h"
+#include "ephy-glib-compat.h"
+#include "ephy-string.h"
 
 #include <time.h>
 #include <string.h>
 #include <glib/gi18n.h>
-#include <libgnomevfs/gnome-vfs-uri.h>
 
-#define EPHY_HISTORY_XML_ROOT	 "ephy_history"
-#define EPHY_HISTORY_XML_VERSION "1.0"
+#define EPHY_HISTORY_XML_ROOT	 (const xmlChar *)"ephy_history"
+#define EPHY_HISTORY_XML_VERSION (const xmlChar *)"1.0"
 
-/* how often to save the history, in milliseconds */
-#define HISTORY_SAVE_INTERVAL (5 * 60 * 1000)
+/* how often to save the history, in seconds */
+#define HISTORY_SAVE_INTERVAL (5 * 60)
 
 /* if you change this remember to change also the user interface description */
 #define HISTORY_PAGE_OBSOLETE_DAYS 10
 
+/* the number of seconds in a day */
+#define SECS_PER_DAY (60*60*24)
+
 #define EPHY_HISTORY_GET_PRIVATE(object)(G_TYPE_INSTANCE_GET_PRIVATE ((object), EPHY_TYPE_HISTORY, EphyHistoryPrivate))
 
-struct EphyHistoryPrivate
+struct _EphyHistoryPrivate
 {
 	char *xml_file;
 	EphyNodeDb *db;
@@ -55,11 +59,17 @@ struct EphyHistoryPrivate
 	EphyNode *last_page;
 	GHashTable *hosts_hash;
 	GHashTable *pages_hash;
-	int autosave_timeout;
+	guint autosave_timeout;
 	guint update_hosts_idle;
+	guint disable_history_notifier_id;
 	gboolean dirty;
 	gboolean enabled;
-	guint disable_history_notifier_id;
+};
+
+enum
+{
+	REDIRECT_FLAG	= 1 << 0,
+	TOPLEVEL_FLAG	= 1 << 1
 };
 
 enum
@@ -70,62 +80,22 @@ enum
 
 enum
 {
+	ADD_PAGE,
 	VISITED,
 	CLEARED,
-        LAST_SIGNAL
+	REDIRECT,
+	ICON_UPDATED,
+	LAST_SIGNAL
 };
 
-static guint signals[LAST_SIGNAL] = { 0 };
+static guint signals[LAST_SIGNAL];
 
 static void ephy_history_class_init	(EphyHistoryClass *klass);
 static void ephy_history_init		(EphyHistory *history);
 static void ephy_history_finalize	(GObject *object);
+static gboolean impl_add_page           (EphyHistory *, const char *, gboolean, gboolean);
 
-static GObjectClass *parent_class = NULL;
-
-GType
-ephy_history_get_type (void)
-{
-        static GType type = 0;
-
-        if (type == 0)
-        {
-                static const GTypeInfo our_info =
-                {
-                        sizeof (EphyHistoryClass),
-                        NULL, /* base_init */
-                        NULL, /* base_finalize */
-                        (GClassInitFunc) ephy_history_class_init,
-                        NULL,
-                        NULL, /* class_data */
-                        sizeof (EphyHistory),
-                        0, /* n_preallocs */
-                        (GInstanceInitFunc) ephy_history_init
-                };
-
-                type = g_type_register_static (G_TYPE_OBJECT,
-					       "EphyHistory",
-					       &our_info, 0);
-        }
-
-        return type;
-}
-
-static void
-ephy_history_set_enabled (EphyHistory *history,
-			  gboolean enabled)
-{
-	LOG ("ephy_history_set_enabled %d", enabled)
-
-	history->priv->enabled = enabled;
-
-	ephy_node_db_set_immutable (history->priv->db, !enabled);
-
-	if (enabled == FALSE)
-	{
-		ephy_history_clear (history);
-	}
-}
+G_DEFINE_TYPE (EphyHistory, ephy_history, G_TYPE_OBJECT)
 
 static void
 ephy_history_set_property (GObject *object,
@@ -156,7 +126,7 @@ ephy_history_get_property (GObject *object,
 		case PROP_ENABLED:
 			g_value_set_boolean (value, history->priv->enabled);
 			break;
-		}
+	}
 }
 
 static void
@@ -164,19 +134,32 @@ ephy_history_class_init (EphyHistoryClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-        parent_class = g_type_class_peek_parent (klass);
-
         object_class->finalize = ephy_history_finalize;
 	object_class->get_property = ephy_history_get_property;
 	object_class->set_property = ephy_history_set_property;
 
+	klass->add_page = impl_add_page;
+
 	g_object_class_install_property (object_class,
 					 PROP_ENABLED,
 					 g_param_spec_boolean ("enabled",
 							       "Enabled",
 							       "Enabled",
 							       TRUE,
-							       G_PARAM_READWRITE));
+							       G_PARAM_READWRITE | G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB));
+
+	signals[ADD_PAGE] =
+		g_signal_new ("add_page",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (EphyHistoryClass, add_page),
+			      g_signal_accumulator_true_handled, NULL,
+			      ephy_marshal_BOOLEAN__STRING_BOOLEAN_BOOLEAN,
+			      G_TYPE_BOOLEAN,
+			      3,
+			      G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE,
+			      G_TYPE_BOOLEAN,
+			      G_TYPE_BOOLEAN);
 
 	signals[VISITED] =
                 g_signal_new ("visited",
@@ -199,23 +182,40 @@ ephy_history_class_init (EphyHistoryClass *klass)
                               G_TYPE_NONE,
                               0);
 
-	g_type_class_add_private (object_class, sizeof(EphyHistoryPrivate));
+	signals[REDIRECT] =
+                g_signal_new ("redirect",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_FIRST,
+                              G_STRUCT_OFFSET (EphyHistoryClass, redirect),
+                              NULL, NULL,
+                              ephy_marshal_VOID__STRING_STRING,
+                              G_TYPE_NONE,
+                              2,
+			      G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE,
+			      G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);
+
+	signals[ICON_UPDATED] =
+		g_signal_new ("icon_updated",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      G_STRUCT_OFFSET (EphyHistoryClass, icon_updated),
+			      NULL, NULL,
+			      ephy_marshal_VOID__STRING_STRING,
+			      G_TYPE_NONE,
+			      2,
+			      G_TYPE_STRING, G_TYPE_STRING);
+
+	g_type_class_add_private (object_class, sizeof (EphyHistoryPrivate));
 }
 
 static gboolean
-page_is_obsolete (EphyNode *node, GDate *now)
+page_is_obsolete (EphyNode *node, time_t now)
 {
 	int last_visit;
-	GDate date;
 
 	last_visit = ephy_node_get_property_int
 		(node, EPHY_NODE_PAGE_PROP_LAST_VISIT);
-
-        g_date_clear (&date, 1);
-        g_date_set_time (&date, last_visit);
-
-	return (g_date_days_between (&date, now) >=
-		HISTORY_PAGE_OBSOLETE_DAYS);
+	return now - last_visit >= HISTORY_PAGE_OBSOLETE_DAYS*SECS_PER_DAY;
 }
 
 static void
@@ -223,12 +223,9 @@ remove_obsolete_pages (EphyHistory *eb)
 {
 	GPtrArray *children;
 	int i;
-	GTime now;
-	GDate current_date;
+	time_t now;
 
 	now = time (NULL);
-        g_date_clear (&current_date, 1);
-        g_date_set_time (&current_date, time (NULL));
 
 	children = ephy_node_get_children (eb->priv->pages);
 	for (i = (int) children->len - 1; i >= 0; i--)
@@ -237,34 +234,41 @@ remove_obsolete_pages (EphyHistory *eb)
 
 		kid = g_ptr_array_index (children, i);
 
-		if (page_is_obsolete (kid, &current_date))
+		if (page_is_obsolete (kid, now))
 		{
 			ephy_node_unref (kid);
 		}
 	}
 }
 
+static gboolean
+save_filter (EphyNode *node,
+	     EphyNode *page_node)
+{
+	return node != page_node;
+}
+
 static void
 ephy_history_save (EphyHistory *eb)
 {
 	int ret;
 
 	/* only save if there are changes */
-	if (eb->priv->dirty == FALSE)
+	if (eb->priv->dirty == FALSE && eb->priv->enabled)
 	{
 		return;
 	}
 
-	LOG ("Saving history db")
+	LOG ("Saving history db");
 
 	ret = ephy_node_db_write_to_xml_safe
-		(eb->priv->db, eb->priv->xml_file,
+		(eb->priv->db, (const xmlChar *)eb->priv->xml_file,
 		 EPHY_HISTORY_XML_ROOT,
 		 EPHY_HISTORY_XML_VERSION,
 		 NULL, /* comment */
 		 eb->priv->hosts,
-		 1, eb->priv->pages,
-		 eb->priv->pages, 0,
+		 (EphyNodeFilterFunc) save_filter, eb->priv->pages,
+		 eb->priv->pages, NULL, NULL,
 		 NULL);
 
 	if (ret >=0)
@@ -377,13 +381,9 @@ update_host_on_child_remove (EphyNode *node)
 
 	if (host_last_visit != new_host_last_visit)
 	{
-		GValue value = { 0, };
-
-		g_value_init (&value, G_TYPE_INT);
-		g_value_set_int (&value, new_host_last_visit);
-		ephy_node_set_property (node, EPHY_NODE_PAGE_PROP_LAST_VISIT,
-				        &value);
-		g_value_unset (&value);
+		ephy_node_set_property_int (node,
+					    EPHY_NODE_PAGE_PROP_LAST_VISIT,
+					    new_host_last_visit);
 	}
 }
 
@@ -435,6 +435,24 @@ page_removed_from_host_cb (EphyNode *node,
 	}
 }
 
+static void
+remove_pages_from_host_cb (EphyNode *host,
+			   EphyHistory *eh)
+{
+	GPtrArray *children;
+	EphyNode *site;
+	int i;
+
+	children = ephy_node_get_children (host);
+
+	for (i = (int) children->len - 1; i >= 0; i--)
+	{
+		site = g_ptr_array_index (children, i);
+
+		ephy_node_unref (site);
+	}
+}
+
 static void
 connect_page_removed_from_host (char *url,
                                 EphyNode *node,
@@ -446,6 +464,10 @@ connect_page_removed_from_host (char *url,
 					 EPHY_NODE_CHILD_REMOVED,
 				         (EphyNodeCallback) page_removed_from_host_cb,
 					 G_OBJECT (eb));
+	ephy_node_signal_connect_object (node,
+					 EPHY_NODE_DESTROY,
+					 (EphyNodeCallback) remove_pages_from_host_cb,
+					 G_OBJECT (eb));
 }
 
 static void
@@ -461,8 +483,8 @@ disable_history_notifier (GConfClient *client,
 static void
 ephy_history_init (EphyHistory *eb)
 {
-	GValue value = { 0, };
 	EphyNodeDb *db;
+	const char *all = _("All");
 
         eb->priv = EPHY_HISTORY_GET_PRIVATE (eb);
 	eb->priv->update_hosts_idle = 0;
@@ -482,21 +504,18 @@ ephy_history_init (EphyHistory *eb)
 
 	/* Pages */
 	eb->priv->pages = ephy_node_new_with_id (db, PAGES_NODE_ID);
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, _("All"));
-	ephy_node_set_property (eb->priv->pages,
-			        EPHY_NODE_PAGE_PROP_LOCATION,
-			        &value);
-	ephy_node_set_property (eb->priv->pages,
-			        EPHY_NODE_PAGE_PROP_TITLE,
-			        &value);
-	g_value_unset (&value);
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, EPHY_NODE_ALL_PRIORITY);
-	ephy_node_set_property (eb->priv->pages,
-			        EPHY_NODE_PAGE_PROP_PRIORITY,
-			        &value);
-	g_value_unset (&value);
+
+	ephy_node_set_property_string (eb->priv->pages,
+				       EPHY_NODE_PAGE_PROP_LOCATION,
+				       all);
+	ephy_node_set_property_string (eb->priv->pages,
+				       EPHY_NODE_PAGE_PROP_TITLE,
+				       all);
+
+	ephy_node_set_property_int (eb->priv->pages,
+				    EPHY_NODE_PAGE_PROP_PRIORITY,
+				    EPHY_NODE_ALL_PRIORITY);
+	
 	ephy_node_signal_connect_object (eb->priv->pages,
 					 EPHY_NODE_CHILD_ADDED,
 				         (EphyNodeCallback) pages_added_cb,
@@ -540,7 +559,7 @@ ephy_history_init (EphyHistory *eb)
 
 	/* setup the periodic history saving callback */
 	eb->priv->autosave_timeout =
-		g_timeout_add (HISTORY_SAVE_INTERVAL,
+		g_timeout_add_seconds (HISTORY_SAVE_INTERVAL,
 		       (GSourceFunc)periodic_save_cb,
 		       eb);
 
@@ -578,7 +597,7 @@ ephy_history_finalize (GObject *object)
 
 	LOG ("Global history finalized");
 
-	G_OBJECT_CLASS (parent_class)->finalize (object);
+	G_OBJECT_CLASS (ephy_history_parent_class)->finalize (object);
 }
 
 EphyHistory *
@@ -592,49 +611,27 @@ ephy_history_host_visited (EphyHistory *eh,
 			   EphyNode *host,
 			   GTime now)
 {
-	GValue value = { 0, };
 	int visits;
 
-	LOG ("Host visited")
+	LOG ("Host visited");
 
 	visits = ephy_node_get_property_int
 		(host, EPHY_NODE_PAGE_PROP_VISITS);
 	if (visits < 0) visits = 0;
 	visits++;
 
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, visits);
-	ephy_node_set_property (host, EPHY_NODE_PAGE_PROP_VISITS,
-			        &value);
-	g_value_unset (&value);
-
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, now);
-	ephy_node_set_property (host, EPHY_NODE_PAGE_PROP_LAST_VISIT,
-			        &value);
-	g_value_unset (&value);
+	ephy_node_set_property_int (host, EPHY_NODE_PAGE_PROP_VISITS, visits);
+	ephy_node_set_property_int (host, EPHY_NODE_PAGE_PROP_LAST_VISIT,
+				    now);
 }
 
 static EphyNode *
-ephy_history_add_host (EphyHistory *eh, EphyNode *page)
+internal_get_host (EphyHistory *eh, const char *url, gboolean create)
 {
-	const char *url;
-
-	url = ephy_node_get_property_string
-		(page, EPHY_NODE_PAGE_PROP_LOCATION);
-
-	return ephy_history_get_host (eh, url);
-}
-
-EphyNode *
-ephy_history_get_host (EphyHistory *eh, const char *url)
-{
-	GnomeVFSURI *vfs_uri = NULL;
 	EphyNode *host = NULL;
-	const char *host_name = NULL;
+	char *host_name = NULL;
 	GList *host_locations = NULL, *l;
-	GValue value = { 0, };
-	const char *scheme = NULL;
+	char *scheme = NULL;
 	GTime now;
 
 	g_return_val_if_fail (url != NULL, NULL);
@@ -646,24 +643,22 @@ ephy_history_get_host (EphyHistory *eh, const char *url)
 
 	now = time (NULL);
 
-	vfs_uri = gnome_vfs_uri_new (url);
-
-	if (vfs_uri)
+	if (url)
 	{
-		scheme = gnome_vfs_uri_get_scheme (vfs_uri);
-		host_name = gnome_vfs_uri_get_host_name (vfs_uri);
+		scheme = g_uri_parse_scheme (url);
+		host_name = ephy_string_get_host_name (url);
 	}
 
 	/* Build an host name */
 	if (scheme == NULL || host_name == NULL)
 	{
-		host_name = _("Others");
+		host_name = g_strdup (_("Others"));
 		host_locations = g_list_append (host_locations,
 						g_strdup ("about:blank"));
 	}
 	else if (strcmp (scheme, "file") == 0)
 	{
-		host_name = _("Local files");
+		host_name = g_strdup (_("Local files"));
 		host_locations = g_list_append (host_locations,
 						g_strdup ("file:///"));
 	}
@@ -671,23 +666,34 @@ ephy_history_get_host (EphyHistory *eh, const char *url)
 	{
 		char *location;
 		char *tmp;
+		
+		if (g_str_equal (scheme, "https"))
+		{
+			/* If scheme is https, we still fake http. */
+			location = g_strconcat ("http://", host_name, "/", NULL);
+			host_locations = g_list_append (host_locations, location);
+		}
 
+		/* We append the real address */
 		location = g_strconcat (scheme,
 					"://", host_name, "/", NULL);
 		host_locations = g_list_append (host_locations, location);
 
-		if (g_str_has_prefix (host_name, "www."))
+		/* and also a fake www-modified address if it's http or https. */
+		if (g_str_has_prefix (scheme, "http"))
 		{
-			tmp = g_strdup (g_utf8_offset_to_pointer (host_name, 4));
+			if (g_str_has_prefix (host_name, "www."))
+			{
+				tmp = g_strdup (host_name + 4);
+			}
+			else
+			{
+				tmp = g_strconcat ("www.", host_name, NULL);
+			}
+			location = g_strconcat ("http://", tmp, "/", NULL);
+			g_free (tmp);
+			host_locations = g_list_append (host_locations, location);
 		}
-		else
-		{
-			tmp = g_strconcat ("www.", host_name, NULL);
-		}
-		location = g_strconcat (gnome_vfs_uri_get_scheme (vfs_uri),
-					"://", tmp, "/", NULL);
-		g_free (tmp);
-		host_locations = g_list_append (host_locations, location);
 	}
 
 	g_return_val_if_fail (host_locations != NULL, NULL);
@@ -699,52 +705,63 @@ ephy_history_get_host (EphyHistory *eh, const char *url)
 		if (host) break;
 	}
 
-	if (!host)
+	if (!host && create)
 	{
 		host = ephy_node_new (eh->priv->db);
 		ephy_node_signal_connect_object (host,
 						 EPHY_NODE_CHILD_REMOVED,
 					         (EphyNodeCallback) page_removed_from_host_cb,
 						 G_OBJECT (eh));
-
-		g_value_init (&value, G_TYPE_STRING);
-		g_value_set_string (&value, host_name);
-		ephy_node_set_property (host, EPHY_NODE_PAGE_PROP_TITLE,
-				        &value);
-		g_value_unset (&value);
-
-		g_value_init (&value, G_TYPE_STRING);
-		g_value_set_string (&value, (char *)host_locations->data);
-		ephy_node_set_property (host, EPHY_NODE_PAGE_PROP_LOCATION,
-				        &value);
-		g_value_unset (&value);
-
-		g_value_init (&value, G_TYPE_INT);
-		g_value_set_int (&value, now);
-		ephy_node_set_property (host, EPHY_NODE_PAGE_PROP_FIRST_VISIT,
-				        &value);
-		g_value_unset (&value);
-
+		ephy_node_signal_connect_object (host,
+						 EPHY_NODE_DESTROY,
+						 (EphyNodeCallback) remove_pages_from_host_cb,
+						 G_OBJECT (eh));
+		ephy_node_set_property_string (host,
+					       EPHY_NODE_PAGE_PROP_TITLE,
+					       host_name);
+		ephy_node_set_property_string (host,
+					       EPHY_NODE_PAGE_PROP_LOCATION,
+					       (char *)host_locations->data);
+		ephy_node_set_property_int (host,
+					    EPHY_NODE_PAGE_PROP_FIRST_VISIT,
+					    now);
 		ephy_node_add_child (eh->priv->hosts, host);
 	}
 
-	ephy_history_host_visited (eh, host, now);
-
-	if (vfs_uri)
+	if (host)
 	{
-		gnome_vfs_uri_unref (vfs_uri);
+		ephy_history_host_visited (eh, host, now);
 	}
 
+	g_free (scheme);
+	g_free (host_name);
+
 	g_list_foreach (host_locations, (GFunc)g_free, NULL);
 	g_list_free (host_locations);
 
 	return host;
 }
 
+EphyNode *
+ephy_history_get_host (EphyHistory *eh, const char *url)
+{
+	return internal_get_host (eh, url, FALSE);
+}
+
+static EphyNode *
+ephy_history_add_host (EphyHistory *eh, EphyNode *page)
+{
+	const char *url;
+
+	url = ephy_node_get_property_string
+		(page, EPHY_NODE_PAGE_PROP_LOCATION);
+
+	return internal_get_host (eh, url, TRUE);
+}
+
 static void
 ephy_history_visited (EphyHistory *eh, EphyNode *node)
 {
-	GValue value = { 0, };
 	GTime now;
 	int visits;
 	const char *url;
@@ -762,22 +779,14 @@ ephy_history_visited (EphyHistory *eh, EphyNode *node)
 	if (visits < 0) visits = 0;
 	visits++;
 
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, visits);
-	ephy_node_set_property (node, EPHY_NODE_PAGE_PROP_VISITS,
-			        &value);
-	g_value_unset (&value);
-
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, now);
-	ephy_node_set_property (node, EPHY_NODE_PAGE_PROP_LAST_VISIT,
-			        &value);
+	ephy_node_set_property_int (node, EPHY_NODE_PAGE_PROP_VISITS, visits);
+	ephy_node_set_property_int (node, EPHY_NODE_PAGE_PROP_LAST_VISIT,
+				    now);
 	if (visits == 1)
 	{
-		ephy_node_set_property
-			(node, EPHY_NODE_PAGE_PROP_FIRST_VISIT, &value);
+		ephy_node_set_property_int
+			(node, EPHY_NODE_PAGE_PROP_FIRST_VISIT, now);
 	}
-	g_value_unset (&value);
 
 	host_id = ephy_node_get_property_int (node, EPHY_NODE_PAGE_PROP_HOST_ID);
 	if (host_id >= 0)
@@ -812,46 +821,60 @@ ephy_history_get_page_visits (EphyHistory *gh,
 }
 
 void
-ephy_history_add_page (EphyHistory *eb,
-		       const char *url)
+ephy_history_add_page (EphyHistory *eh,
+		       const char *url,
+		       gboolean redirect,
+		       gboolean toplevel)
+{
+	gboolean result = FALSE;
+
+	g_signal_emit (eh, signals[ADD_PAGE], 0, url, redirect, toplevel, &result);
+}
+
+static gboolean
+impl_add_page (EphyHistory *eb,
+	       const char *url,
+	       gboolean redirect,
+	       gboolean toplevel)
 {
 	EphyNode *bm, *node, *host;
-	GValue value = { 0, };
+	gulong flags = 0;
 
 	if (eb->priv->enabled == FALSE)
 	{
-		return;
+		return FALSE;
 	}
 
 	node = ephy_history_get_page (eb, url);
 	if (node)
 	{
 		ephy_history_visited (eb, node);
-		return;
+		return TRUE;
 	}
 
 	bm = ephy_node_new (eb->priv->db);
 
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, url);
-	ephy_node_set_property (bm, EPHY_NODE_PAGE_PROP_LOCATION,
-			        &value);
-	ephy_node_set_property (bm, EPHY_NODE_PAGE_PROP_TITLE,
-			        &value);
-	g_value_unset (&value);
+	ephy_node_set_property_string (bm, EPHY_NODE_PAGE_PROP_LOCATION, url);
+	ephy_node_set_property_string (bm, EPHY_NODE_PAGE_PROP_TITLE, url);
+
+	if (redirect) flags |= REDIRECT_FLAG;
+	if (toplevel) flags |= TOPLEVEL_FLAG;
+
+	/* EphyNode SUCKS! */
+	ephy_node_set_property_long (bm, EPHY_NODE_PAGE_PROP_EXTRA_FLAGS,
+				     flags);
 
 	host = ephy_history_add_host (eb, bm);
 
-	g_value_init (&value, G_TYPE_INT);
-	g_value_set_int (&value, ephy_node_get_id (host));
-	ephy_node_set_property (bm, EPHY_NODE_PAGE_PROP_HOST_ID,
-			        &value);
-	g_value_unset (&value);
+	ephy_node_set_property_int (bm, EPHY_NODE_PAGE_PROP_HOST_ID,
+				    ephy_node_get_id (host));
 
 	ephy_history_visited (eb, bm);
 
 	ephy_node_add_child (host, bm);
 	ephy_node_add_child (eb->priv->pages, bm);
+
+	return TRUE;
 }
 
 EphyNode *
@@ -878,20 +901,16 @@ ephy_history_set_page_title (EphyHistory *gh,
 			     const char *title)
 {
 	EphyNode *node;
-	GValue value = { 0, };
 
-	LOG ("Set page title")
+	LOG ("Set page title");
 
 	if (title == NULL || title[0] == '\0') return;
 
 	node = ephy_history_get_page (gh, url);
 	if (node == NULL) return;
 
-	g_value_init (&value, G_TYPE_STRING);
-	g_value_set_string (&value, title);
-	ephy_node_set_property
-		(node, EPHY_NODE_PAGE_PROP_TITLE, &value);
-	g_value_unset (&value);
+	ephy_node_set_property_string (node, EPHY_NODE_PAGE_PROP_TITLE,
+				       title);
 }
 
 const char*
@@ -919,21 +938,23 @@ ephy_history_set_icon (EphyHistory *gh,
 		       const char *url,
 		       const char *icon)
 {
-	EphyNode *host;
+	EphyNode *node, *host;
+	int host_id;
 
-	LOG ("Set host icon")
+	node = ephy_history_get_page (gh, url);
+	if (node == NULL) return;
+	
+	host_id = ephy_node_get_property_int (node, EPHY_NODE_PAGE_PROP_HOST_ID);
+	g_return_if_fail (host_id >= 0);
 
-	host = g_hash_table_lookup (gh->priv->hosts_hash, url);
+	host = ephy_node_db_get_node_from_id (gh->priv->db, host_id);
 	if (host)
 	{
-		GValue value = { 0, };
-
-		g_value_init (&value, G_TYPE_STRING);
-		g_value_set_string (&value, icon);
-		ephy_node_set_property
-			(host, EPHY_NODE_PAGE_PROP_ICON, &value);
-		g_value_unset (&value);
+		ephy_node_set_property_string (host, EPHY_NODE_PAGE_PROP_ICON,
+					       icon);
 	}
+
+	g_signal_emit (gh, signals[ICON_UPDATED], 0, url, icon);
 }
 
 void
@@ -941,7 +962,7 @@ ephy_history_clear (EphyHistory *gh)
 {
 	EphyNode *node;
 
-	LOG ("clearing history")
+	LOG ("clearing history");
 
 	ephy_node_db_set_immutable (gh->priv->db, FALSE);
 
@@ -984,3 +1005,39 @@ ephy_history_is_enabled (EphyHistory *history)
 
 	return history->priv->enabled;
 }
+
+void
+ephy_history_set_enabled (EphyHistory *history,
+			  gboolean enabled)
+{
+	int ret;
+
+	ret = 1;
+
+	LOG ("ephy_history_set_enabled %d", enabled);
+
+	/* Write history only when disabling it, not when reenabling it */
+	if (!enabled && history->priv->dirty)
+	{
+		ret = ephy_node_db_write_to_xml_safe
+			(history->priv->db, (const xmlChar *)history->priv->xml_file,
+			 EPHY_HISTORY_XML_ROOT,
+			 EPHY_HISTORY_XML_VERSION,
+			 NULL, /* comment */
+			 history->priv->hosts,
+			 (EphyNodeFilterFunc) save_filter, history->priv->pages,
+			 history->priv->pages, NULL, NULL,
+			 NULL);
+	}
+
+	if (ret >=0)
+	{
+		/* save was successful */
+		history->priv->dirty = FALSE;
+	}
+
+	history->priv->enabled = enabled;
+
+	ephy_node_db_set_immutable (history->priv->db, !enabled);
+}
+
